/*
 * Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <gst/gst.h>
#include <glib.h>

#include <math.h>

#include <stdio.h>
#include <string.h>
#include "cuda_runtime_api.h"


#include "gstnvdsmeta.h" 
#include "gstnvdsinfer.h"

#define PGIE_CONFIG_FILE  "dstensor_sgie1_config.txt"
#define SGIE1_CONFIG_FILE "dstensor_sgie2_config.txt"
#define SGIE2_CONFIG_FILE "dstensor_sgie3_config.txt"

#define MUXER_OUTPUT_WIDTH 1280
#define MUXER_OUTPUT_HEIGHT 720
#define MUXER_BATCH_TIMEOUT_USEC 4000000

static GstPadProbeReturn
queue_pad_buffer_probe (GstPad * pad, GstPadProbeInfo * info, gpointer u_data)
{
  static uint use_device_mem = 0;
  NvDsBatchMeta *batch_meta =
      gst_buffer_get_nvds_batch_meta (GST_BUFFER (info->data));

  /* Iterate each frame metadata in batch */
  for (NvDsMetaList * l_frame = batch_meta->frame_meta_list; l_frame != NULL;
      l_frame = l_frame->next) {
    NvDsFrameMeta *frame_meta = (NvDsFrameMeta *) l_frame->data;

    uint infer_count = 0;
    /* Iterate user metadata in frames to search PGIE's tensor metadata */
    for (NvDsMetaList * l_user = frame_meta->frame_user_meta_list;
        l_user != NULL; l_user = l_user->next) {
      NvDsUserMeta *user_meta = (NvDsUserMeta *) l_user->data;
      if (user_meta->base_meta.meta_type != NVDSINFER_TENSOR_OUTPUT_META)
        continue;

      char infer_result[512];
      snprintf(infer_result, 32, "Primary_infer_index:%d ", infer_count++);
      /* convert to tensor metadata */
      NvDsInferTensorMeta *meta =
          (NvDsInferTensorMeta *) user_meta->user_meta_data;
      for (unsigned int i = 0; i < meta->num_output_layers; i++) {
        NvDsInferLayerInfo *info = &meta->output_layers_info[i];
        info->buffer = meta->out_buf_ptrs_host[i];
        if (use_device_mem) {
          cudaMemcpy (meta->out_buf_ptrs_host[i], meta->out_buf_ptrs_dev[i],
              info->dims.numElements * 4, cudaMemcpyDeviceToHost);
        }
      }
      
      NvDsInferDimsCHW dims;
      getDimsCHWFromDims (dims, meta->output_layers_info[0].dims);
      uint numClasses = dims.c != 1 ? dims.c : (dims.h != 1 ? dims.h :dims.w);
      auto *outputCoverageBuffer = (float *) meta->output_layers_info[0].buffer;
      for (unsigned int i = 0; i < numClasses; i++){
        snprintf(infer_result + strlen(infer_result), 32, "class%d:%f", i, outputCoverageBuffer[i]);
      }
      g_print("%s\n", infer_result);
    }
  }
  use_device_mem = 1 - use_device_mem;
  return GST_PAD_PROBE_OK;
}

static GstPadProbeReturn
sink_pad_buffer_probe (GstPad * pad, GstPadProbeInfo * info, gpointer u_data)
{
  static uint use_device_mem = 0;
  NvDsBatchMeta *batch_meta =
      gst_buffer_get_nvds_batch_meta (GST_BUFFER (info->data));

  /* Iterate each frame metadata in batch */
  for (NvDsMetaList * l_frame = batch_meta->frame_meta_list; l_frame != NULL;
      l_frame = l_frame->next) {
    NvDsFrameMeta *frame_meta = (NvDsFrameMeta *) l_frame->data;

    uint infer_count = 0;
    /* Iterate user metadata in frames to search PGIE's tensor metadata */
    for (NvDsMetaList * l_user = frame_meta->frame_user_meta_list;
        l_user != NULL; l_user = l_user->next) {
      NvDsUserMeta *user_meta = (NvDsUserMeta *) l_user->data;
      if (user_meta->base_meta.meta_type != NVDSINFER_TENSOR_OUTPUT_META)
        continue;

      char infer_result[512];
      snprintf(infer_result, 32, "All_infer_index:%d ", infer_count++);
      /* convert to tensor metadata */
      NvDsInferTensorMeta *meta =
          (NvDsInferTensorMeta *) user_meta->user_meta_data;
      for (unsigned int i = 0; i < meta->num_output_layers; i++) {
        NvDsInferLayerInfo *info = &meta->output_layers_info[i];
        info->buffer = meta->out_buf_ptrs_host[i];
        if (use_device_mem) {
          cudaMemcpy (meta->out_buf_ptrs_host[i], meta->out_buf_ptrs_dev[i],
              info->dims.numElements * 4, cudaMemcpyDeviceToHost);
        }
      }
      
      NvDsInferDimsCHW dims;
      getDimsCHWFromDims (dims, meta->output_layers_info[0].dims);
      uint numClasses = dims.c != 1 ? dims.c : (dims.h != 1 ? dims.h :dims.w);
      auto *outputCoverageBuffer = (float *) meta->output_layers_info[0].buffer;
      for (unsigned int i = 0; i < numClasses; i++){
        snprintf(infer_result + strlen(infer_result), 32, "class%d:%f", i, outputCoverageBuffer[i]);
      }
      g_print("%s\n", infer_result);
    }
  }
  use_device_mem = 1 - use_device_mem;
  return GST_PAD_PROBE_OK;
}

static gboolean
bus_call (GstBus * bus, GstMessage * msg, gpointer data)
{
  GMainLoop *loop = (GMainLoop *) data;
  switch (GST_MESSAGE_TYPE (msg)) {
    case GST_MESSAGE_EOS:
      g_print ("End of stream\n");
      g_main_loop_quit (loop);
      break;
    case GST_MESSAGE_ERROR:{
      gchar *debug;
      GError *error;
      gst_message_parse_error (msg, &error, &debug);
      g_printerr ("ERROR from element %s: %s\n",
          GST_OBJECT_NAME (msg->src), error->message);
      if (debug)
        g_printerr ("Error details: %s\n", debug);
      g_free (debug);
      g_error_free (error);
      g_main_loop_quit (loop);
      break;
    }
    default:
      break;
  }
  return TRUE;
}

int
main (int argc, char *argv[])
{
  GMainLoop *loop = NULL;
  GstElement *pipeline = NULL, *source = NULL, *h264parser = NULL, *decoder = NULL, *streammux = NULL, 
      *pgie = NULL, *queue = NULL, *sgie1 = NULL, *sgie2 = NULL, *sink = NULL;

  GstBus *bus = NULL;
  guint bus_watch_id = 0;
  GstPad *queue_pad = NULL, *sink_pad = NULL;
  guint i;

  /* Check input arguments */
  if (argc < 2) {
    g_printerr
        ("Usage: %s <elementary H264 file 1> ... <elementary H264 file n>\n",
        argv[0]);
    return -1;
  }

  guint num_sources = argc - 1;

  /* Standard GStreamer initialization */
  gst_init (&argc, &argv);
  loop = g_main_loop_new (NULL, FALSE);

  /* Create gstreamer elements */

  /* Create Pipeline element that will be a container of other elements */
  pipeline = gst_pipeline_new ("dstensor-pipeline");


  /* Create nvstreammux instance to form batches from one or more sources. */
  streammux = gst_element_factory_make ("nvstreammux", "stream-muxer");

  if (!pipeline || !streammux) {
    g_printerr ("One element could not be created. Exiting.\n");
    return -1;
  }

  pgie = gst_element_factory_make ("nvinfer", "primary-nvinference-engine");
  
  queue = gst_element_factory_make ("queue", "queue");

  sgie1 = gst_element_factory_make ("nvinfer", "secondary1-nvinference-engine");

  sgie2 = gst_element_factory_make ("nvinfer", "secondary2-nvinference-engine");

  sink = gst_element_factory_make ("fakesink", "fakesink");

  if (!pgie || !queue || !sgie1 || !sgie2 || !sink) {
    g_printerr ("One element could not be created. Exiting.\n");
    return -1;
  }

  g_object_set (G_OBJECT (streammux), "width", MUXER_OUTPUT_WIDTH, "height",
      MUXER_OUTPUT_HEIGHT, "batch-size", num_sources,
      "batched-push-timeout", MUXER_BATCH_TIMEOUT_USEC, NULL);

  g_object_set (G_OBJECT (pgie), "config-file-path", PGIE_CONFIG_FILE,
      "output-tensor-meta", TRUE, "interval", 1, "batch-size", num_sources, NULL);
  g_object_set (G_OBJECT (sgie1), "config-file-path", SGIE1_CONFIG_FILE,
      "output-tensor-meta", TRUE, "process-mode", 1, NULL);
  g_object_set (G_OBJECT (sgie2), "config-file-path", SGIE2_CONFIG_FILE,
      "output-tensor-meta", TRUE, "process-mode", 1, NULL);

  /* we add a message handler */
  bus = gst_pipeline_get_bus (GST_PIPELINE (pipeline));
  bus_watch_id = gst_bus_add_watch (bus, bus_call, loop);
  gst_object_unref (bus);

  gst_bin_add_many (GST_BIN (pipeline),
      streammux, pgie, queue, sgie1, sgie2, sink, NULL);

  for (i = 0; i < num_sources; i++) {
    /* Source element for reading from the file */
    source = gst_element_factory_make ("filesrc", NULL);

    /* Since the data format in the input file is elementary h264 stream,
     * we need a h264parser */
    h264parser = gst_element_factory_make ("h264parse", NULL);

    /* Use nvdec_h264 for hardware accelerated decode on GPU */
    decoder = gst_element_factory_make ("nvv4l2decoder", NULL);
    gst_bin_add_many (GST_BIN (pipeline), source, h264parser, decoder, NULL);

    if (!source || !h264parser || !decoder) {
      g_printerr ("One element could not be created. Exiting.\n");
      return -1;
    }

    GstPad *sinkpad, *srcpad;
    gchar pad_name_sink[16];
    sprintf (pad_name_sink, "sink_%d", i);
    gchar pad_name_src[16] = "src";

    sinkpad = gst_element_get_request_pad (streammux, pad_name_sink);
    if (!sinkpad) {
      g_printerr ("Streammux request sink pad failed. Exiting.\n");
      return -1;
    }

    srcpad = gst_element_get_static_pad (decoder, pad_name_src);
    if (!srcpad) {
      g_printerr ("Decoder request src pad failed. Exiting.\n");
      return -1;
    }

    if (gst_pad_link (srcpad, sinkpad) != GST_PAD_LINK_OK) {
      g_printerr ("Failed to link decoder to stream muxer. Exiting.\n");
      return -1;
    }

    gst_object_unref (sinkpad);
    gst_object_unref (srcpad);

    /* Link the elements together */
    if (!gst_element_link_many (source, h264parser, decoder, NULL)) {
      g_printerr ("Elements could not be linked: 1. Exiting.\n");
      return -1;
    }
    /* Set the input filename to the source element */
    g_object_set (G_OBJECT (source), "location", argv[i + 1], NULL);
  }

  if (!gst_element_link_many (streammux, pgie, queue, sgie1, sgie2, sink, NULL)) {
    g_printerr ("Elements could not be linked. Exiting.\n");
    return -1;
  }
  
/*  queue_pad = gst_element_get_static_pad (queue, "sink");
  if (!queue_pad)
    g_print ("Unable to get queue pad\n");
  else
    gst_pad_add_probe (queue_pad, GST_PAD_PROBE_TYPE_BUFFER,
        queue_pad_buffer_probe, NULL, NULL);
*/        
  sink_pad = gst_element_get_static_pad (sink, "sink");
  if (!sink_pad)
    g_print ("Unable to get sink pad\n");
  else
    gst_pad_add_probe (sink_pad, GST_PAD_PROBE_TYPE_BUFFER,
        sink_pad_buffer_probe, NULL, NULL);

  GST_DEBUG_BIN_TO_DOT_FILE_WITH_TS (GST_BIN (pipeline),
      GST_DEBUG_GRAPH_SHOW_ALL, "dsi-app-null");
  /* Set the pipeline to "playing" state */
  g_print ("Now playing: %s\n", argv[1]);
  gst_element_set_state (pipeline, GST_STATE_PLAYING);

  /* Iterate */
  g_print ("Running...\n");
  g_main_loop_run (loop);

  /* Out of the main loop, clean up nicely */
  g_print ("Returned, stopping playback\n");
  gst_element_set_state (pipeline, GST_STATE_NULL);
  g_print ("Deleting pipeline\n");
  gst_object_unref (GST_OBJECT (pipeline));
  g_source_remove (bus_watch_id);
  g_main_loop_unref (loop);
  return 0;
}
